#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>

#include "protein/atoms_reading.h"
#include "protein/atoms_classification.h"
#include "protein/residue_ids_collection.h"

#include "auxiliaries/command_line_options.h"
#include "auxiliaries/file_header.h"
#include "auxiliaries/vector_io.h"
#include "auxiliaries/map_io.h"

#include "resources/vdwr.h"

protein::VanDerWaalsRadiusAssigner construct_radius_assigner(const std::string& radius_classes_file_name, const std::string& radius_members_file_name)
{
	std::istream* radius_classes_stream=0;
	if(radius_classes_file_name.empty())
	{
		radius_classes_stream=new std::istringstream(std::string(reinterpret_cast<const char*>(vdwr_classes), vdwr_classes_len));
	}
	else
	{
		radius_classes_stream=new std::ifstream(radius_classes_file_name.c_str());
	}

	std::istream* radius_members_stream=0;
	if(radius_members_file_name.empty())
	{
		radius_members_stream=new std::istringstream(std::string(reinterpret_cast<const char*>(vdwr_members), vdwr_members_len));
	}
	else
	{
		radius_members_stream=new std::ifstream(radius_members_file_name.c_str());
	}

	const protein::VanDerWaalsRadiusAssigner radius_assigner(*radius_classes_stream, *radius_members_stream);

	delete radius_classes_stream;
	delete radius_members_stream;

	return radius_assigner;
}

void collect_atoms(const auxiliaries::CommandLineOptions& clo)
{
	clo.check_allowed_options("--radius-classes: --radius-members: --HETATM --HOH --no-classification --rename-chain:");

	std::string radius_classes_file_name="";
	std::string radius_members_file_name="";
	if(clo.isopt("--radius-classes") || clo.isopt("--radius-members"))
	{
		radius_classes_file_name=clo.arg<std::string>("--radius-classes");
		radius_members_file_name=clo.arg<std::string>("--radius-members");
	}

	const bool include_heteroatoms=clo.isopt("--HETATM");
	const bool include_water=clo.isopt("--HOH");
	const std::string chain_renaming=clo.isopt("--rename-chain") ? clo.arg<std::string>("--rename-chain") : std::string("");

	const protein::VanDerWaalsRadiusAssigner radius_assigner=construct_radius_assigner(radius_classes_file_name, radius_members_file_name);

	std::vector<protein::Atom> atoms=protein::AtomsReading::read_atoms_from_PDB_file_stream(std::cin, radius_assigner, include_heteroatoms, include_water);

	protein::AtomsClassification::classify_atoms(atoms);

	if(!chain_renaming.empty())
	{
		for(std::size_t i=0;i<atoms.size();i++)
		{
			atoms[i].chain_id=chain_renaming;
		}
	}

	if(atoms.empty())
	{
		throw std::runtime_error("No atoms were collected from the provided PDB file stream");
	}
	else
	{
		auxiliaries::print_vector(std::cout, "atoms", atoms);
	}
}

void collect_residue_ids(const auxiliaries::CommandLineOptions& clo)
{
	clo.check_allowed_options("");

	const std::vector<protein::Atom> atoms=auxiliaries::read_vector<protein::Atom>(std::cin, "atoms", "atoms", false);

	const std::map<protein::ResidueID, protein::ResidueSummary> residue_ids=protein::collect_residue_ids_from_atoms(atoms);

	if(residue_ids.empty())
	{
		throw std::runtime_error("No residue identifiers were collected from the provided atoms stream");
	}
	else
	{
		auxiliaries::print_map(std::cout, "residue_ids", residue_ids, false);
	}
}

void merge_atoms(const auxiliaries::CommandLineOptions& clo)
{
	clo.check_allowed_options("");
	std::vector<protein::Atom> atoms=auxiliaries::read_vector<protein::Atom>(std::cin, "atoms", "atoms", true);
	while(auxiliaries::check_file_header(std::cin, "atoms"))
	{
		std::vector<protein::Atom> more_atoms=auxiliaries::read_vector<protein::Atom>(std::cin, "more atoms", "", true);
		atoms.insert(atoms.end(), more_atoms.begin(), more_atoms.end());
	}
	auxiliaries::print_vector(std::cout, "atoms", atoms);
}
